#Licensed Materials - Property of IBM 
#Rational ClearQuest 
#Copyright IBM Corp. 2008, 2010 All Rights Reserved 

package cqcli::newquery;

use strict;

require Exporter;

use Error qw(:try);
use cqcli::Exception;
use Inline::Java qw(caught cast) ;
use cqcli::Util;
use cqcli::IOHandler;
use cqcli::commonservice;
use Text::ParseWords;
use Getopt::Long 2.34;


our @ISA = qw(Exporter);

our %EXPORT_TAGS = ( 'all' => [ qw() ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw(newquery);

our $VERSION = '0.01';

#######################################################################
our @ARGVBAACKUP = @ARGV;
our %RepetitionCount = ();
our $isLoginOpt = undef;
our $CliSession = undef;
our $JavaReqObj = undef;
our %CmdInfo = ( 'cmdstring' => [ ], 'cmdname'  => "new_query");
our %CmdLogin = ('database'	=>	undef,'schema'=>  undef,
			 'user'		=>	undef,'password'=>  undef);
our $BLANK_MARKER_SET = 0;
our $Replace = 0;
our	$TypeDef = "Defect";
our	$Displayfields = "";
our $FieldName = undef;
our $FilterFields = undef;
our @filterTokens = undef;
our $QueryName = undef;

our %DisplayConfiguration = ();


#########################################################################

#########handle getopt warnings begin########################

sub InitParser
{
	my @tmpARGV = @_;
	@ARGVBAACKUP = @ARGV;
	@ARGV = undef;
	$isLoginOpt = undef;
	%RepetitionCount = ();
	%CmdInfo = ('cmdstring' => [ ], 'cmdname'  => "new_query");
	%CmdLogin =('database'	=>undef,'schema'=>  undef,
				'user'	=>	undef,'password'=>  undef);

	%DisplayConfiguration = ();

	$Replace = 0;
	
	$TypeDef = "Defect";
	$Displayfields = "";
	$FieldName = undef;
	$FilterFields = "";
	@filterTokens = undef;
	$QueryName = undef;
	
	$CmdInfo{'cmdstring'} =[@tmpARGV];
	($CmdInfo{'cmdname'}  , @ARGV ) = @{$CmdInfo{'cmdstring'}};
	$CliSession = GetSession();
	$JavaReqObj = GetJavaHash();
	return 1;
}

sub CleanUpParser()
{

	@ARGV = @ARGVBAACKUP;
	%RepetitionCount = ();
	%CmdInfo = ('cmdstring' => [ ], 'cmdname'  => "new_query");
	%CmdLogin =('database'	=>undef,'schema'=>  undef,
				'user'	=>	undef,'password'=>  undef);

	$Replace = 0;	
	
	%DisplayConfiguration = ();
	
	$TypeDef = "Defect";
	$Displayfields = "";
	$FieldName = undef;
	$FilterFields = "";
	@filterTokens = undef;
	$QueryName = undef;
	
	$CmdInfo{'cmdstring'} = undef;
	$CmdInfo{'cmdname'} =undef;
	$CliSession = undef;
	$JavaReqObj = undef;

}

#########handle getopt warnings end#####################################

sub handle_help_usage
{
	my $opt = $_[0];
	
	if($opt eq "help")	{
		ThrowUsageException($CmdInfo{'cmdname'},$opt);
	}
	else{
		ThrowUsageException($CmdInfo{'cmdname'},$opt);
	}
	return 1;
}

sub handle_ask_mode
{
	my $mode = shift;
	SetCQCLIAskMode(1);
	return 1;
}


sub handle_login_data
{
	my ($opt , $val) = @_;
	my $msg = undef;
	$isLoginOpt = 1;

	if($RepetitionCount{$opt}) {
		ThrowSyntaxErrorException("CQCLI_Repeat_Option","-$opt", $CmdInfo{'cmdname'} );
	}else {
		$RepetitionCount{$opt} = 1;
	}

	if( ($_[0] eq "database") && ($val)	){
		$CmdLogin{'database'} = $val;
	}elsif(($_[0] eq "user") && ($val)) {
		$CmdLogin{'user'} = $val;
	}elsif(($_[0] eq "password") && ($val)) {
		$CmdLogin{'password'} = $val;
	}elsif(($_[0] eq "password") && ($val eq "")){
		$CmdLogin{'password'} = "";
		$BLANK_MARKER_SET = 1 ;
	}
	elsif((($_[0] eq "schema") || ($_[0] eq "master")) && ($val) ) {
		$CmdLogin{'schema'} = $val; 
	}else{
		ThrowSyntaxErrorException("CQCLI_Missing_Value","-$opt",$CmdInfo{'cmdname'});
	}
	return 1;
}

sub handle_recordtype_list
{
	my ($opt , $type) = @_;
	my $msg = undef;
	if($RepetitionCount{'$opt'} ) {
		ThrowSyntaxErrorException("CQCLI_Repeat_Option","-$opt", $CmdInfo{'cmdname'} );
	}else {
		$RepetitionCount{'$opt'} = 1;
	}

	if($type){
		$TypeDef = $type;
	}else {
		ThrowSyntaxErrorException("CQCLI_Missing_Value","-$opt", $CmdInfo{'cmdname'} );
	}
	return 1;
}

sub handle_field_list
{
	my ($opt , $val) = @_;
	my $msg = undef;

	if( ($val)	){
		$Displayfields = $val . ',' . $Displayfields ;
	}else{
		ThrowSyntaxErrorException("CQCLI_Missing_Value","-$opt",$CmdInfo{'cmdname'});
	}

	return 1;
}

sub handle_replace_field
{
	my ($opt ) = shift;
	my $msg = undef;

	if($RepetitionCount{'handle_replace_field'}) {
		ThrowSyntaxErrorException("CQCLI_Repeat_Action","[$opt ]", $CmdInfo{'cmdname'} );
	}else {
		$RepetitionCount{'handle_replace_field'} = 1;
	}
	
	$Replace = 1;
	
	return 1;
}

sub handle_display_options
{
	my ($opt) = @_;
	
	#all_fields|af|sort_names|sn|type_names|tn|value|va|no_recname|nr|no_headers|nh|no_names|nn|no_values|nv|no_dup_info|nd
		if($opt eq "s" || $opt eq "show" ) {
			$DisplayConfiguration{'SHOW'} = 1;
		}
		elsif($opt eq "show_sql" || $opt eq "ss" ) {
			$DisplayConfiguration{'SHOW_SQL'} = 1;
		}
		elsif($opt eq "show_fields" || $opt eq "sf" ) {
			$DisplayConfiguration{'SHOW_FIELDS'} = 1;
		}
		elsif($opt eq "show_filter" || $opt eq "sfi" ) {
			$DisplayConfiguration{'SHOW_FILTER'} = 1;
		}
		elsif($opt eq "no_recname" || $opt eq "nr" ) {
			$DisplayConfiguration{'NO_RECORD_HEADER'} = 1;
		}
		elsif($opt eq "dbids") {
			$DisplayConfiguration{'SHOW_DBID'} = 1;
		}
		else {
			ThrowSyntaxErrorException("CQCLI_Missing_Value","Display field config error", $CmdInfo{'cmdname'} );
		}
	
	return 1;
}

sub parseForHelpUsage()
{
	my $ret = undef;

	my $helpRules = {'help|h|?'		=>	\&handle_help_usage	,
					 'man'			=>	\&handle_help_usage	};
	
	Getopt::Long::Configure('autoabbrev','ignore_case','permute','pass_through');
	$ret = GetOptions( %{$helpRules} );
	Getopt::Long::Configure('default');

	return 1;
}

sub parseForCoreOptions()
{
	my $ret = undef;
	my $CoreOptRules = {
			'type:s'   =>	\&handle_recordtype_list, 
			'field:s'  =>   \&handle_field_list,
			'replace|rep:s' => \&handle_replace_field,
			's|show' => \&handle_display_options,
			'show_sql|ss' => \&handle_display_options,
			'show_fields|sf' => \&handle_display_options,
			'show_filter|sfi' => \&handle_display_options,
			'dbids' => \&handle_display_options
	};
	Getopt::Long::Configure( 'autoabbrev','ignore_case','permute','pass_through');
	$ret = GetOptions( %{$CoreOptRules} );
	Getopt::Long::Configure('default');

	return 1;
}

sub parseForLoginCredentials 
{
	my $ret = undef;
	my $CommonRules = {
		'database|dbs|d:s'					=>  \&handle_login_data			,
		'user|login|u|l:s'					=>	\&handle_login_data			,
		'schema|s|m|master:s'				=>	\&handle_login_data			,
		'password|p:s'						=>	\&handle_login_data			,
		'type:s'							=>	\&handle_recordtype_list			,
	};
	Getopt::Long::Configure( 'autoabbrev','ignore_case','permute','pass_through');
	$ret = GetOptions( %{$CommonRules} );
	Getopt::Long::Configure('default');
	return 1;
}

sub createBinaryToken
{
	my ($operand1,$operator,$operand2) = @_;
	my $filterstring = undef ;
	
	
	$filterstring = $operand1 . "[" . $operator . "]" . $operand2 ;
	
	return $filterstring;
	
}

sub createBinaryTokenForRangeFields
{
	my ($operand1,$operator,$operand2) = @_;
	my @operands = split(' ',$operand2);
	my $filterstring = undef;
	
	$filterstring = $operand1 . "[" . $operator . "]" . $operands[0] . "[and]" . $operands[1] ;
	
	return $filterstring;
}

sub createVariadicTokenForRangeFields
{
	my ($operand1,$operator,$operand2) = @_;
	my @operands = split(' ',$operand2);
	my $filterstring = undef;
	
	$filterstring = $operand1 . "[" . $operator . "]";
	
	foreach my $ops (@operands) {
	
		$filterstring = $filterstring . $ops . "[or]" ;
	}
	
	$filterstring =  substr($filterstring, 0, - 4);
	
	return $filterstring;
}

sub createNullTokenFields
{
	my ($operand1,$operator) = @_;
	my $filterstring = undef;
	
      $filterstring = $operand1 . "[" . $operator . "]" ;

	return $filterstring;
}

sub createFilterNode
{
	my ($leftleaf,$operator,$rightleaf) = @_;
	my $filterNode;
	
	$filterNode = "[(]"  . "," . $leftleaf . ","  . "[" . $operator . "]"  . "," . $rightleaf . "," . "[)]" ;
	return $filterNode;
}

sub createLeftNode
{
	my($filterNode,$operator) = @_;

		push(@filterTokens,"[(]");
		push(@filterTokens,$filterNode);
		push(@filterTokens,$operator);
}

sub createRightNode
{
	my($filterNode) = @_;
	push(@filterTokens,$filterNode);
	push(@filterTokens,"[)]");
	
	$FilterFields = join(',',@filterTokens);
	@filterTokens = undef;
}

sub parseForUnknownOptions
{
	my @tmpARGV = @ARGV;
	my $lmt = $#tmpARGV;
	my $msg = undef;
	my $leftLeafToBeSet = 0;
	my $rightLeafToBeSet = 0;
	my $leftleaf = undef;
	my $rightleaf = undef;
	my $filterString = "";
	my @operandStack = undef;
	my $poppedOperand = undef;
	my $whichNode = 0;
	
	
	if($lmt == -1) {
		ThrowSyntaxErrorException("CQCLI_Missing_Value","<query filters> and <display fields>",$CmdInfo{'cmdname'});
	}

	for (my $cnt = 0; $cnt <= $lmt;$cnt++)  
	{	

		if( $tmpARGV[$cnt] =~ /^(-)((eq)|(ne)|(lt)|(le)|(gt)|(ge)|(prompt)|(like)|(not_like))/i)
		{
			$filterString = $filterString . createBinaryToken($tmpARGV[$cnt+1],$2,$tmpARGV[$cnt+2]) . ",";
			$cnt = $cnt + 2;
		}
		elsif( $tmpARGV[$cnt] =~ /^(-)((between)|(not_between))/i )
		{
			$filterString = $filterString . createBinaryTokenForRangeFields($tmpARGV[$cnt+1],$2,$tmpARGV[$cnt+2]) . ",";
			$cnt = $cnt + 2;
		}
		elsif( $tmpARGV[$cnt] =~ /^(-)((in)|(not_in))/i )
		{
			$filterString = $filterString . createVariadicTokenForRangeFields($tmpARGV[$cnt+1],$2,$tmpARGV[$cnt+2]) . ",";
			$cnt = $cnt + 2;
		}
		elsif( $tmpARGV[$cnt] =~ /^(-)((null)|(not_null))/i )
		{
			$filterString = $filterString . createNullTokenFields($tmpARGV[$cnt+1],$2) . ",";
			$cnt = $cnt + 1;
		}
		elsif($tmpARGV[$cnt] =~ /^(-)((and)|(or))/i)
		{
			$filterString = 	$filterString . "[" . $2 . "]" . ",";
		}
		elsif($tmpARGV[$cnt] =~ /^(\()/)
		{
			$filterString = 	$filterString . "[" . "(" . "]" . ",";
		}
		elsif($tmpARGV[$cnt] =~ /^(\))/)
		{
			$filterString = 	$filterString . "[" . ")" . "]" . ",";
		}
		elsif($tmpARGV[$cnt] =~ /^(-)(.+)/)
		{
			my $fname = $2;
			$Displayfields	= $Displayfields . $fname . ",";
			splice @ARGV,0,2;
			next;
		}
		else
		{
			$QueryName = $tmpARGV[$cnt];
		}

	}

	$FilterFields = $filterString;
	return 1;
}

sub SetDisplayConfig()
{
	#set request object for successfull field validation
	my $displayhash = GetJavaHash();
	my $fld = undef;
	foreach $fld (keys %DisplayConfiguration) {
		$displayhash->put($fld,$DisplayConfiguration{$fld})
	}
	$JavaReqObj->put("control",$displayhash);
}

sub Parse()
{
	try {

		parseForHelpUsage();
		parseForLoginCredentials();
		parseForCoreOptions();
		parseForUnknownOptions();
		
		if(!($QueryName))
		{
				$JavaReqObj->put("save",undef);
		}
		else {
				$JavaReqObj->put("save",$QueryName);
		}
				
		if($Displayfields eq "")
		{
				my $promptmsg = GetLocalizedMessage('CQCLI_Warn_DisplayFields');
				CLIPrintOutput($promptmsg);
		}
				
		$JavaReqObj->put("type",$TypeDef);	
		if($FilterFields) {
			$JavaReqObj->put("filter",$FilterFields);
		}
		
		if($Replace eq 1){
			$JavaReqObj->put("name",$QueryName);
		}


		
	}
	catch cqcli::Exception with {
		my $ex = shift;	
		throw $ex;
	}
	otherwise {
		if (caught("java.lang.Exception")){
			ThrowJavaException($@,"CQCLI_ErrJava_ParserUnknown");
		}
		else{
			ThrowParseException('CQCLI_Err_ParserUnknown',$@);
		}
    }
	finally {
		Getopt::Long::Configure('default');
	};

	return 1;
}

sub ValidateSessionInfo()
{
	
	if($isLoginOpt){
		ValidateLoginCredentials();
		return;
	}elsif($CliSession){
		return;
	}else{
		ValidateLoginCredentials();
		return;
	}

}
sub ValidateLoginCredentials()
{
	if(!($CmdLogin{'user'})){
		$CmdLogin{'user'} = GetUserName();
	}
	if(!($CmdLogin{'password'}) && $BLANK_MARKER_SET == 0){
		$CmdLogin{'password'} = GetPassword();
	}
	if(!($CmdLogin{'database'})){
		$CmdLogin{'database'} = GetDBName();
	}
	if(!($CmdLogin{'schema'})){
		$CmdLogin{'schema'} = GetSchemaName();
	}
	$CliSession=CreateSession(%CmdLogin);
	# is not a command scope session
	SetSessionScope(0);
}

sub ValidateFieldNames()
{
	my @fieldtobevalidated = undef;
	my $recordtypeobj = GetRecordTypeInfoObj($TypeDef);
	my $expectedfield = undef ;
	
	if($Displayfields ne ""){
		@fieldtobevalidated = split(',',$Displayfields)
	}
	else {
		return 1;
	}

	$Displayfields = "";

	my $length = scalar @fieldtobevalidated;
	for( my $i=0; $i<$length;$i++) {
		$expectedfield = $recordtypeobj->isValidFieldName($fieldtobevalidated[$i]);

		if($expectedfield) {
			if($expectedfield =~ /\,/ ){
				ThrowValidationException('CQCLI_Failure_FieldAmbiguity',$fieldtobevalidated[$i],$expectedfield,$TypeDef);
			}
			else {
				$Displayfields = $Displayfields . $expectedfield . ",";
			}
		}
		else {
				ThrowValidationException("CQCLI_Invalid_FieldInput",$TypeDef);
		}
	}

		
		$JavaReqObj->put("display",$Displayfields);
	
}

sub Validate()
{
	try {
		ValidateSessionInfo();
		ValidateFieldNames();
	}
	catch cqcli::Exception with {
		my $ex = shift;		
		throw $ex;
	}
	otherwise {
		if (caught("java.lang.Exception")){
			ThrowJavaException($@,"CQCLI_ErrJava_ValidationUnknown");
		}
		else{
			ThrowValidationException('CQCLI_Err_ValidationUnknown',$@);
		}		
    }
	finally {	};

	return 1;
}

sub Set()
{
	SetDisplayConfig();
}

sub newquery
{
	my @tmpCmd = @_;
	my $req = undef;
	try {
		InitParser(@tmpCmd);
		Parse();
		Validate();
		Set();
		$req =  GetJavaPayload("new_query",$JavaReqObj);

		SetRequestObject($req);
		return 1;
	}finally {
		CleanUpParser();
	};
}

1;